from inspect import Parameter, isawaitable, signature
from typing import Awaitable, TypeVar


def appender(to_append: list):
    def append_processer(func):
        to_append.append(func)

        return func

    return append_processer


def is_positional_parameter(p: Parameter):
    return p.kind is Parameter.POSITIONAL_OR_KEYWORD or p.kind is Parameter.KEYWORD_ONLY


def count_position_parameters(func):
    return sum(map(is_positional_parameter, signature(func).parameters.values()))


T = TypeVar("T")


async def resolve(maybe_awaitable: T | Awaitable[T], /) -> T:
    while isawaitable(maybe_awaitable):
        maybe_awaitable = await maybe_awaitable

    return maybe_awaitable  # type: ignore



from sys import version_info
from typing import Literal

from .utils import is_message_start

Role = Literal["user", "assistant", "system"]

if version_info >= (3, 11):
    from typing import NotRequired, TypedDict

    class Message(TypedDict):  # type: ignore
        role: Role
        content: str
        name: NotRequired[str]

else:
    from typing_extensions import NotRequired, TypedDict

    class Message(TypedDict):
        role: Role
        content: str
        name: NotRequired[str]


class MessageBuilder:
    _initializing = True
    __slots__ = ("role", "content", "name")

    def __init__(self, role: Role, /, content: str = "", name: str | None = None):
        self.role: Role = role
        self.content = content
        self.name = name

    def __repr__(self):
        if self.name is not None:
            return f"<| {self.role} {self.name} |>"
        return f"<| {self.role} |>"

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __setattr__(self, key, value) -> None:
        if not self._initializing:
            assert self is not U and self is not A and self is not S
            assert isinstance(value, str)
        return super().__setattr__(key, value)

    def __matmul__(self, name: str):
        assert isinstance(name, str) and name
        return self.__class__(self.role, self.content, name)

    def dict(self) -> Message:
        if self.name:
            return {"role": self.role, "content": self.content, "name": self.name}
        return {"role": self.role, "content": self.content}

    def __gt__(self, content: str) -> Message:
        assert isinstance(content, str)
        if self.name:
            return {"role": self.role, "content": content, "name": self.name}
        return {"role": self.role, "content": content}


U = user = MessageBuilder("user")
A = assistant = MessageBuilder("assistant")
S = system = MessageBuilder("system")
MessageBuilder._initializing = False


def ensure(text_or_list: list[Message] | str) -> list[Message]:
    return (
        parse_chat_markup(text_or_list)
        if isinstance(text_or_list, str)
        else text_or_list
    )


def parse_chat_markup(text: str) -> list[Message]:
    messages = []
    current_message = None
    buffer = []

    for line in text.splitlines():
        match = is_message_start.match(line)
        if match:
            role, name = match.group(1), match.group(2)

            if current_message:
                current_message["content"] = "\n".join(buffer)
                messages.append(current_message)
                buffer.clear()

            current_message = {"role": role, "content": ""}

            if name:
                current_message["name"] = name

        elif current_message:
            buffer.append(line)

    if current_message:
        current_message["content"] = "\n".join(buffer)
        messages.append(current_message)

    return messages or [{"role": "user", "content": text}]



from functools import cached_property
from typing import Any, Callable, ParamSpec, TypeVar

from openai import AsyncOpenAI, OpenAI  # type: ignore

from ...prompt.chat import Message, ensure
from ..base import *

P = ParamSpec("P")
T = TypeVar("T")


def with_params(_: Callable[P, Any]):
    """provide type hints with no runtime overhead"""

    def inherit(_: type[T]) -> Callable[..., Callable[P, T]]:
        return lambda as_is: as_is

    return inherit


inherit = with_params(OpenAI)


class Config(Configurable):
    @cached_property
    def client(self):
        return OpenAI(**self._config)

    @cached_property
    def aclient(self):
        return AsyncOpenAI(**self._config)


@inherit(Complete)
class TextComplete(Config):
    def __call__(self, text: str, /, **config):
        config = config | {"stream": False, "prompt": text}
        result = self.client.completions.create(**config)
        return result.choices[0].text


@inherit(AsyncComplete)
class AsyncTextComplete(Config):
    async def __call__(self, text: str, /, **config):
        config = config | {"stream": False, "prompt": text}
        result = await self.aclient.completions.create(**config)
        return result.choices[0].text


@inherit(Generate)
class TextGenerate(Config):
    def __call__(self, text: str, /, **config):
        config = config | {"stream": True, "prompt": text}
        stream = self.client.completions.create(**config)
        for event in stream:
            yield event.choices[0].text


@inherit(AsyncGenerate)
class AsyncTextGenerate(Config):
    async def __call__(self, text: str, /, **config):
        config = config | {"stream": True, "prompt": text}
        stream = await self.aclient.completions.create(**config)
        async for event in stream:
            yield event.choices[0].text


@inherit(Complete)
class ChatComplete(Config):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = config | {"stream": False, "messages": messages}
        result = self.client.chat.completions.create(**config)
        return result.choices[0].message.content


@inherit(AsyncComplete)
class AsyncChatComplete(Config):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = config | {"stream": False, "messages": messages}
        result = await self.aclient.chat.completions.create(**config)
        return result.choices[0].message.content


@inherit(Generate)
class ChatGenerate(Config):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = config | {"stream": True, "messages": messages}
        stream = self.client.chat.completions.create(**config)
        for event in stream:
            yield event.choices[0].delta.content or ""


@inherit(AsyncGenerate)
class AsyncChatGenerate(Config):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = config | {"stream": True, "messages": messages}
        stream = await self.aclient.chat.completions.create(**config)
        async for event in stream:
            yield event.choices[0].delta.content or ""
